import numpy as np
from numpy import save

from timeit import default_timer as timer
import math
import sys


def resourceAllocator(orderList, rewardList):
    maximumSubset = []  # list for storing the index of the accepted orders
    result = np.zeros((orderList.shape[0],), dtype=int)  # array for storing the maximum total reward of each episode

    # setting all elements of the dp array to 0
    for episode in range(orderList.shape[0]):
        numOfOrder = orderList.shape[1]
        # dp stores the different states of the resources
        dp = []
        for i in range(numOfOrder + 1):
            # 6 loops, representing the different states of the resource consumption
            list6 = []
            for j in range(7):
                list5 = []
                for k in range(7):
                    list4 = []
                    for l in range(7):
                        list3 = []
                        for m in range(7):
                            list2 = []
                            for n in range(7):
                                list1 = []
                                for o in range(7):
                                    list1.append(0)
                                list2.append(list1)
                            list3.append(list2)
                        list4.append(list3)
                    list5.append(list4)
                list6.append(list5)
            dp.append(list6)

        # i represents the order
        for i in range(numOfOrder):
            # different states, repesenting the number of the consumption of each resource (6 types)
            for j in range(1, 7):
                for k in range(1, 7):
                    for l in range(1, 7):
                        for m in range(1, 7):
                            for n in range(1, 7):
                                for o in range(1, 7):
                                    # ignoring the current request, i
                                    if i > 0:
                                        dp[i + 1][j][k][l][m][n][o] = dp[i][j][k][l][m][n][o]

                                    # accepting the current request, i if it's possible to accept
                                    if orderList[episode, i, 0] <= j and orderList[episode, i, 1] <= k and orderList[
                                        episode, i, 2] <= l and orderList[episode, i, 3] <= m and orderList[
                                        episode, i, 4] <= n and orderList[episode, i, 5] <= o:
                                        # compare accept order or not accept order
                                        dp[i + 1][j][k][l][m][n][o] = max(dp[i + 1][j][k][l][m][n][o],  # if not accept
                                                                          rewardList[
                                                                              episode, i, 1] +  # if accept, plus current reward, the substract the resources
                                                                          dp[i][j - orderList[episode, i, 0]][
                                                                              k - orderList[episode, i, 1]][
                                                                              l - orderList[episode, i, 2]][
                                                                              m - orderList[episode, i, 3]][
                                                                              n - orderList[episode, i, 4]][
                                                                              o - orderList[episode, i, 5]])

        result[episode] = dp[numOfOrder][6][6][6][6][6][6]  # maximum reward for this episode

        # find the index of the accepted orders
        idx = numOfOrder - 1
        cur = result[episode]
        rem = [6, 6, 6, 6, 6, 6]
        subset = []

        while cur > 0:
            if dp[idx][rem[0]][rem[1]][rem[2]][rem[3]][rem[4]][rem[5]] == cur:  # ignoring the current request, i
                idx -= 1
            else:  # accepting the current request, i
                subset.append(idx)  # adding the request index to the subset
                rem[0] -= orderList[episode, idx, 0]
                rem[1] -= orderList[episode, idx, 1]
                rem[2] -= orderList[episode, idx, 2]
                rem[3] -= orderList[episode, idx, 3]
                rem[4] -= orderList[episode, idx, 4]
                rem[5] -= orderList[episode, idx, 5]
                cur -= rewardList[episode, idx, 1]
                idx -= 1
        maximumSubset.append(subset)

    return result, maximumSubset


def generateGame(NUM_EPISODES, NUM_ORDERS, NUM_SPACEWIDTH, NUM_CAPACITY, REWARD_MIN_PER_UNIT=1, REWARD_MAX_PER_UNIT=9):
    base_element_set = np.array(
        # Data based on Wendtris Java code (Hagen)
        [[1, 2, 3, 2, 1, 0],
         [1, 1, 1, 1, 0, 0],
         [0, 2, 2, 0, 0, 0],
         [2, 2, 1, 1, 0, 0],
         [1, 3, 2, 2, 1, 0],
         [2, 1, 1, 1, 3, 3]], dtype=object)

    ARR_ORDERLIST = np.zeros((NUM_EPISODES, NUM_ORDERS, NUM_SPACEWIDTH), dtype=np.int8)

    ARR_SET_AND_POSITION = np.zeros((NUM_EPISODES, NUM_ORDERS, 3), dtype=np.int8)
    # 0 -> element (index of base_element_set)
    # 1 -> position (offset)
    # 2 -> length of element

    ARR_REWARDS = np.zeros((NUM_EPISODES, NUM_ORDERS, 2), dtype=np.float32)
    # 0 -> reward per unit
    # 1 -> total reward

    # Index 0: Set index
    ARR_SET_AND_POSITION[:, :, 0] = np.random.randint(base_element_set.shape[0], size=(NUM_EPISODES, NUM_ORDERS))

    # Set reward data per unit
    ARR_REWARDS[:, :, 0] = np.random.randint(REWARD_MIN_PER_UNIT, REWARD_MAX_PER_UNIT + 1,
                                             size=(NUM_EPISODES, NUM_ORDERS))

    # Index 1: Position of element
    for i in range(NUM_EPISODES):
        for j in range(NUM_ORDERS):
            # Set radon position for the elements
            len_element = len(base_element_set[ARR_SET_AND_POSITION[i, j, 0]])
            pos_element = np.random.randint(NUM_SPACEWIDTH - len_element + 1)
            ARR_SET_AND_POSITION[i, j, 1] = pos_element
            ARR_SET_AND_POSITION[i, j, 2] = len_element

            # set reward data for total element
            ARR_REWARDS[i, j, 1] = ARR_REWARDS[i, j, 0] * sum(base_element_set[ARR_SET_AND_POSITION[i, j, 0]])

            for k in range(len_element):
                ARR_ORDERLIST[i, j, pos_element + k] = base_element_set[ARR_SET_AND_POSITION[i, j, 0]][k]

    return ARR_ORDERLIST, ARR_REWARDS

if __name__ == "__main__":
    np.set_printoptions(threshold=sys.maxsize)

    num_episodes = 1000
    num_orders = 20
    num_spacewidth = 6
    num_capacity = 6

    orders, rewards = generateGame(num_episodes, num_orders, num_spacewidth, num_capacity)

    start = timer()
    result, maximumSubset = resourceAllocator(orders, rewards)
    result = np.array(result)
    maximumSubset = np.array(maximumSubset)
    print(result)
    print(maximumSubset)
    save('order_list.npy', orders)
    save('reward_list.npy', rewards)
    save('results.npy', result)
    save('subset.npy', maximumSubset)
    print("without GPU:", timer()-start)

    print("--------------------------------")