import numpy as np


def buildWendtrisModel():
    global ARR_ALLPOSSIBLEREQUEST_PROB
    global ARR_ALLPOSSIBLEREQUEST
    avg_reward = 5

    ARR_ALLPOSSIBLEREQUEST = np.array(
        # Data based on Wendtris Java code (Hagen)
        [[1, 2, 3, 2, 1, 0],
         [1, 1, 1, 1, 0, 0],
         [0, 2, 2, 0, 0, 0],
         [2, 2, 1, 1, 0, 0],
         [1, 3, 2, 2, 1, 0],
         [2, 1, 1, 1, 3, 3]])
    ARR_ALLPOSSIBLEREQUEST_PROB = np.ones((6, 2), np.float32)
    # 0-> probability for each order
    # 1-> EV for each request: Units * AVG reward * probability for each request

    # calculated "manually":

    # num of different requests, each probability 1/6
    ARR_ALLPOSSIBLEREQUEST_PROB[:, 0] /= 6

    # ressource consumption, hardcoded
    ARR_ALLPOSSIBLEREQUEST_PROB[:, 1] = [9, 4, 4, 6, 9, 11]

    ARR_ALLPOSSIBLEREQUEST_PROB[:, 1] *= avg_reward / 6


buildWendtrisModel()

print(ARR_ALLPOSSIBLEREQUEST_PROB)
print(ARR_ALLPOSSIBLEREQUEST)

sdp_fix_mem = {}


def solveGame_sdp_fixedPosition(cap_remaining: np.array, stage: int):
    global REWARD_AVG_PER_UNIT
    global REWARD_MAX_PER_UNIT
    global ARR_ALLPOSSIBLEREQUEST_PROB
    global ARR_ALLPOSSIBLEREQUEST
    num_req_types = ARR_ALLPOSSIBLEREQUEST_PROB.shape[0]

    cap_remaining_key = cap_remaining.tobytes()
    cap_remaining_total = np.sum(cap_remaining)

    try:
        res_sdp = sdp_fix_mem[stage][cap_remaining_total][cap_remaining_key]
    except KeyError:
        # Check if parent key (cap_remaining_total) exists and create if needed
        try:
            tmp2 = sdp_fix_mem[stage][cap_remaining_total]
        except KeyError:
            # Check if parent key (stage) exists and create if needed
            try:
                tmp1 = sdp_fix_mem[stage]
            except KeyError:
                # create key for state
                sdp_fix_mem[stage] = {}

            # create key for cap_remaining_total
            sdp_fix_mem[stage][cap_remaining_total] = {}

        # create key for order and calculate value
        sdp_fix_mem[stage][cap_remaining_total][cap_remaining_key] = 0.0

        # if we are in stage 1, the value is the mean of the
        # expected rev for all orders which we can and will accept
        if (stage == 1):
            for i in range(num_req_types):
                if np.min(cap_remaining - ARR_ALLPOSSIBLEREQUEST[i]) >= 0:
                    sdp_fix_mem[stage][cap_remaining_total][cap_remaining_key] += ARR_ALLPOSSIBLEREQUEST_PROB[i][1]

        # if we are not in stage 1, we calculate the threshold price for each order type
        # based on the value of the previous stage
        # the value of the current stage at the given capacity vector is the average revenue for each policy
        else:
            sdp_fix_mem[stage][cap_remaining_total][cap_remaining_key] = 0.0
            thres_price = np.ones(num_req_types, np.float32)
            thres_price_ceil = np.ones(num_req_types, np.float32)

            # Calculate V*(cap,stage-1)
            rej_val = solveGame_sdp_fixedPosition(cap_remaining, stage - 1)

            for i in range(num_req_types):
                tmp_cap_remaining = cap_remaining - ARR_ALLPOSSIBLEREQUEST[i]

                # check for request if it can be accepted
                if np.min(tmp_cap_remaining) >= 0:
                    # order can be accepted, calculate V* of stage-1 in case for the remaning capacity
                    tmp_Vcapremaining = solveGame_sdp_fixedPosition(tmp_cap_remaining, stage - 1)

                    # Calculate threshold price and round values to suit the problem with discrete rewards
                    tmp_sum = np.sum(ARR_ALLPOSSIBLEREQUEST[i])
                    thres_price[i] = (rej_val - tmp_Vcapremaining) / tmp_sum
                    if thres_price[i] > 10:
                        thres_price[i] = 10.0

                    thres_price_ceil[i] = np.ceil(thres_price[i])
                    if thres_price_ceil[i] < 1.0:
                        thres_price_ceil[i] = 1.0

                    # Formula for Acceptance probability for given threshold price and discrete rewards
                    acc_prob = (REWARD_MAX_PER_UNIT - thres_price_ceil[i] + 1) / REWARD_MAX_PER_UNIT

                    # In case the order will be accepted, the EV is the average of all discrete values >= threshold
                    acc_EV_per_unit = (thres_price_ceil[i] + REWARD_MAX_PER_UNIT) / 2
                    add_val_acc = acc_prob * ARR_ALLPOSSIBLEREQUEST_PROB[i][0] * (
                                acc_EV_per_unit * tmp_sum + tmp_Vcapremaining)

                    # Prorated value and probability in case threshold price is not met
                    add_val_rej = (1 - acc_prob) * ARR_ALLPOSSIBLEREQUEST_PROB[i][0] * rej_val


                else:
                    # order cant be accepted
                    thres_price[i] = REWARD_MAX_PER_UNIT + 1
                    thres_price_ceil[i] = REWARD_MAX_PER_UNIT + 1
                    # 100% rej probability
                    add_val_rej = ARR_ALLPOSSIBLEREQUEST_PROB[i][0] * rej_val
                    add_val_acc = 0

                # Update Value in Dictionary
                sdp_fix_mem[stage][cap_remaining_total][cap_remaining_key] += add_val_acc + add_val_rej
                print("stage: " + str(stage) + ", rem cap: " + str(cap_remaining) + " thres price ceil: " + str(
                    thres_price_ceil))
            print("Calculated Total Val: " + str(sdp_fix_mem[stage][cap_remaining_total][cap_remaining_key]))

    return sdp_fix_mem[stage][cap_remaining_total][cap_remaining_key]


test = np.array([6, 6, 6, 6, 6, 6])
order = np.array([0, 2, 2, 0, 0, 0])
after_order = np.array([2, 2, 2, 5, 2, 3])
testb = test.tobytes()
after_orderb = after_order.tobytes()
solveGame_sdp_fixedPosition(test, 50)

test = np.array([6, 6, 6, 6, 6, 6])
order = np.array([0, 2, 2, 0, 0, 0])
after_order = np.array([2, 2, 2, 5, 2, 3])
testb = test.tobytes()
after_orderb = after_order.tobytes()
solveGame_sdp_fixedPosition(test, 500)
