import numpy as np
import pandas as pd


class RequestGenerator:

    def __init__(self, num_orders, num_spacewidth=6, num_capacity=6, reward_min_per_unit=1.0,
                 reward_max_per_unit=10.0):
        self.num_orders = num_orders
        self.num_spacewidth = num_spacewidth
        self.num_capacity = num_capacity
        self.reward_min_per_unit = reward_min_per_unit
        self.reward_max_per_unit = reward_max_per_unit

    def generate_requests(self, num_games, seed=None):
        """
        Generate requests
        :param num_games: the number of generated episodes
        :param seed: seed
        :return: order_list, reward_list
        """
        np.random.seed(seed)
        base_element_set = np.array(
            # Data based on Wendtris Java code (Hagen)
            [[1, 2, 3, 2, 1, 0],
             [1, 1, 1, 1, 0, 0],
             [0, 2, 2, 0, 0, 0],
             [2, 2, 1, 1, 0, 0],
             [1, 3, 2, 2, 1, 0],
             [2, 1, 1, 1, 3, 3]], dtype=object)

        arr_orderlist = np.zeros((num_games, self.num_orders, self.num_spacewidth), dtype=np.int8)

        arr_set_and_position = np.zeros((num_games, self.num_orders, 3), dtype=np.int8)
        # 0 -> element (index of base_element_set)
        # 1 -> position (offset)
        # 2 -> length of element

        arr_rewards = np.zeros((num_games, self.num_orders, 2), dtype=np.float16)
        # 0 -> reward per unit
        # 1 -> total reward

        # Index 0: Set index
        arr_set_and_position[:, :, 0] = np.random.randint(base_element_set.shape[0],
                                                          size=(num_games, self.num_orders))

        # Set reward data per unit
        arr_rewards[:, :, 0] = np.random.randint(self.reward_min_per_unit, self.reward_max_per_unit + 1,
                                                 size=(num_games, self.num_orders))

        # Index 1: Position of element
        for i in range(num_games):
            for j in range(self.num_orders):
                # Set radon position for the elements
                len_element = len(base_element_set[arr_set_and_position[i, j, 0]])
                pos_element = np.random.randint(self.num_spacewidth - len_element + 1)
                arr_set_and_position[i, j, 1] = pos_element
                arr_set_and_position[i, j, 2] = len_element

                # set reward data for total element
                arr_rewards[i, j, 1] = arr_rewards[i, j, 0] * sum(base_element_set[arr_set_and_position[i, j, 0]])

                for k in range(len_element):
                    arr_orderlist[i, j, pos_element + k] = base_element_set[arr_set_and_position[i, j, 0]][k]

        return arr_orderlist, arr_rewards
